"""
Detection of the event causes, based on the resource state.

The low-level watch-events are highly limited in information on what
caused them, and they only notify that the object was changed somehow:

* ``ADDED`` for the newly created objects (or for the first-time listing).
* ``MODIFIED`` for the changes of any field, be that metadata, spec, or status.
* ``DELETED`` for the actual deletion of the object post-factum.

The conversion of low-level *events* to high level *causes* is done by
checking the object's state and comparing it to the saved last-seen state.

This allows to track which specific fields were changed, and if are those
changes are important enough to call the handlers: e.g. the ``status`` changes
are ignored, so as some selected system fields of the ``metadata``.

For deletion, the cause is detected when the object is just marked for deletion,
not when it is actually deleted (as the events notify): so that the handlers
could execute on the yet-existing object (and its children, if created).
"""
import dataclasses
import enum
from collections.abc import Mapping
from typing import Any

from kopf._cogs.configs import configuration
from kopf._cogs.structs import bodies, diffs, ephemera, finalizers, \
                               ids, patches, references, reviews
from kopf._core.actions import execution
from kopf._core.intents import stoppers


class Activity(str, enum.Enum):
    STARTUP = 'startup'
    CLEANUP = 'cleanup'
    AUTHENTICATION = 'authentication'
    PROBE = 'probe'


class WebhookType(str, enum.Enum):
    VALIDATING = 'validating'
    MUTATING = 'mutating'

    def __str__(self) -> str:
        return str(self.value)


# Constants for cause types, to prevent a direct usage of strings, and typos.
# They are not exposed by the framework, but are used internally. See also: `kopf.on`.
class Reason(str, enum.Enum):
    CREATE = 'create'
    UPDATE = 'update'
    DELETE = 'delete'
    RESUME = 'resume'
    NOOP = 'noop'
    FREE = 'free'
    GONE = 'gone'

    def __str__(self) -> str:
        return str(self.value)


# These sets are checked in a few places, so we keep them centralised:
# the user-facing causes (for handlers) and internally facing (for the reactor).
HANDLER_REASONS = (
    Reason.CREATE,
    Reason.UPDATE,
    Reason.DELETE,
    Reason.RESUME,
)
REACTOR_REASONS = (
    Reason.NOOP,
    Reason.FREE,
    Reason.GONE,
)
ALL_REASONS = HANDLER_REASONS + REACTOR_REASONS

# The human-readable names of these causes. Will be capitalised when needed.
TITLES = {
    Reason.CREATE.value: 'creation',
    Reason.UPDATE.value: 'updating',
    Reason.DELETE.value: 'deletion',
    Reason.RESUME.value: 'resuming',
}


@dataclasses.dataclass
class BaseCause(execution.Cause):
    """
    Base non-specific cause as used in the framework's reactor in most cases.

    IMPORTANT! Indices overwrite any other kwargs, even the existing ones.

    Why so? Here is why: for forwards & backwards compatibility.
    If an handler uses an index named "children", and Kopf introduces
    a new kwarg "children", the handler's code could break on the upgrade.
    To prevent this, Kopf overwrites the framework's kwarg "children"
    with the operator's index "children" and lets the developers rename it
    when (and if) they want the new kwarg.

    Naming the new indices the same as the known/existing kwargs
    harms only the developers who do so, so this is considered safe.
    """
    indices: ephemera.Indices
    memo: ephemera.AnyMemo

    @property
    def _kwargs(self) -> Mapping[str, Any]:
        kwargs = dict(super()._kwargs)
        del kwargs['indices']
        return kwargs

    @property
    def _super_kwargs(self) -> Mapping[str, Any]:
        return self.indices


@dataclasses.dataclass
class ActivityCause(BaseCause):
    activity: Activity
    settings: configuration.OperatorSettings


@dataclasses.dataclass
class ResourceCause(BaseCause):
    resource: references.Resource
    patch: patches.Patch
    body: bodies.Body

    @property
    def _kwargs(self) -> Mapping[str, Any]:
        return dict(
            super()._kwargs,
            spec=self.body.spec,
            meta=self.body.metadata,
            status=self.body.status,
            uid=self.body.metadata.uid,
            name=self.body.metadata.name,
            namespace=self.body.metadata.namespace,
            labels=self.body.metadata.labels,
            annotations=self.body.metadata.annotations,
        )


@dataclasses.dataclass
class WebhookCause(ResourceCause):
    dryrun: bool
    reason: WebhookType | None  # None means "all" or expects the webhook id
    webhook: ids.HandlerId | None  # None means "all"
    headers: Mapping[str, str]
    sslpeer: Mapping[str, Any]
    userinfo: reviews.UserInfo
    warnings: list[str]  # mutable!
    operation: reviews.Operation | None  # None if not provided for some reason
    subresource: str | None  # e.g. "status", "scale"; None for the main resource body
    old: bodies.Body | None = None
    new: bodies.Body | None = None
    diff: diffs.Diff | None = None

    @property
    def _kwargs(self) -> Mapping[str, Any]:
        kwargs = dict(super()._kwargs)
        del kwargs['reason']
        del kwargs['webhook']
        return kwargs


@dataclasses.dataclass
class IndexingCause(ResourceCause):
    """
    The raw event received from the API.
    """
    pass


@dataclasses.dataclass
class WatchingCause(ResourceCause):
    """
    The raw event received from the API.

    It is a read-only mapping with some extra properties and methods.
    """
    type: bodies.RawEventType
    event: bodies.RawEvent


@dataclasses.dataclass
class SpawningCause(ResourceCause):
    """
    An internal daemon is spawning: tasks, threads, timers.

    Used only on the first appearance of a resource as a container for resource-
    specific objects (loggers, etc).
    """
    reset: bool

    @property
    def _kwargs(self) -> Mapping[str, Any]:
        kwargs = dict(super()._kwargs)
        del kwargs['reset']
        return kwargs


@dataclasses.dataclass
class ChangingCause(ResourceCause):
    """
    The cause is what has caused the whole reaction as a chain of handlers.

    Unlike the low-level Kubernetes watch-events, the cause is aware
    of actual field changes, including multi-handler changes.
    """
    initial: bool
    reason: Reason
    diff: diffs.Diff = diffs.EMPTY
    old: bodies.BodyEssence | None = None
    new: bodies.BodyEssence | None = None

    @property
    def _kwargs(self) -> Mapping[str, Any]:
        kwargs = dict(super()._kwargs)
        del kwargs['initial']
        return kwargs

    @property
    def deleted(self) -> bool:
        """ Used to conditionally skip/select the @on.resume handlers if the object is deleted. """
        return finalizers.is_deletion_ongoing(self.body)


@dataclasses.dataclass
class DaemonCause(ResourceCause):
    """
    An exceptional case of a container for daemon invocation kwargs.

    Regular causes are usually short-term, triggered by a watch-stream event,
    and disappear once the event is processed. The processing includes
    daemon spawning: the original cause and its temporary watch-event
    should not be remembered though the whole life cycle of a daemon.

    Instead, a new artificial daemon-cause is used (this class), which
    passes the kwarg values to the invocation routines. It only contains
    the long-living kwargs: loggers, per-daemon stoppers, body-views
    (with only the latest bodies as contained values), etc.

    Unlike other causes, it is created not in the processing routines once
    per event, but in the daemon spawning routines once per daemon (or a timer).
    Therefore, it is not "detected", but is created directly as an instance.
    """
    stopper: stoppers.DaemonStopper  # a signaller for the termination and its reason.

    @property
    def _kwargs(self) -> Mapping[str, Any]:
        kwargs = dict(super()._kwargs)
        del kwargs['stopper']
        return kwargs

    @property
    def _sync_kwargs(self) -> Mapping[str, Any]:
        return dict(super()._sync_kwargs, stopped=self.stopper.sync_waiter)

    @property
    def _async_kwargs(self) -> Mapping[str, Any]:
        return dict(super()._async_kwargs, stopped=self.stopper.async_waiter)


def detect_watching_cause(
        raw_event: bodies.RawEvent,
        body: bodies.Body,
        **kwargs: Any,
) -> WatchingCause:
    return WatchingCause(
        event=raw_event,
        type=raw_event['type'],
        body=body,
        **kwargs)


def detect_spawning_cause(
        body: bodies.Body,
        **kwargs: Any,
) -> SpawningCause:
    return SpawningCause(
        body=body,
        **kwargs)


def detect_changing_cause(
        *,
        finalizer: str,
        raw_event: bodies.RawEvent,
        body: bodies.Body,
        old: bodies.BodyEssence | None = None,
        new: bodies.BodyEssence | None = None,
        diff: diffs.Diff | None = None,
        initial: bool = False,
        **kwargs: Any,
) -> ChangingCause:
    """
    Detect the cause of the event to be handled.

    This is a purely computational function with no side-effects.
    The causes are then consumed by :func:`process_resource_causes`,
    which performs the actual handler invocation, logging, patching,
    and other side-effects.
    """

    # Put them back to the pass-through kwargs (to avoid code duplication).
    kwargs |= dict(body=body, old=old, new=new, initial=initial)
    if diff is not None:
        kwargs |= dict(diff=diff)

    # The object was really deleted from the cluster. But we do not care anymore.
    if raw_event['type'] == 'DELETED':
        return ChangingCause(reason=Reason.GONE, **kwargs)

    # The finalizer has been just removed. We are fully done.
    deletion_is_ongoing = finalizers.is_deletion_ongoing(body=body)
    deletion_is_blocked = finalizers.is_deletion_blocked(body=body, finalizer=finalizer)
    if deletion_is_ongoing and not deletion_is_blocked:
        return ChangingCause(reason=Reason.FREE, **kwargs)

    if deletion_is_ongoing:
        return ChangingCause(reason=Reason.DELETE, **kwargs)

    # For an object seen for the first time (i.e. just-created), call the creation handlers,
    # then mark the state as if it was seen when the creation has finished.
    # Creation never mixes with resuming, even if an object is detected on startup (first listing).
    if old is None:  # i.e. we have no essence stored
        kwargs['initial'] = False
        return ChangingCause(reason=Reason.CREATE, **kwargs)

    # Cases with no essence changes are usually ignored (NOOP). But for the not-yet-resumed objects,
    # we simulate a fake cause to invoke the resuming handlers. For cases with the essence changes,
    # the resuming handlers will be mixed-in to the regular cause handling ("cuckoo-style")
    # due to the ``initial=True`` flag on the cause, regardless of the reason.
    if not diff and initial:
        return ChangingCause(reason=Reason.RESUME, **kwargs)

    # The previous step triggers one more patch operation without actual changes. Ignore it.
    # Either the last-seen state or the status field has changed.
    if not diff:
        return ChangingCause(reason=Reason.NOOP, **kwargs)

    # And what is left, is the update operation on one of the useful fields of the existing object.
    return ChangingCause(reason=Reason.UPDATE, **kwargs)



import asyncio

import pytest

from kopf._cogs.aiokits.aiotime import sleep


async def test_the_only_delay_is_awaited(looptime):
    unslept = await sleep(123)
    assert looptime == 123
    assert unslept is None


async def test_the_shortest_delay_is_awaited(looptime):
    unslept = await sleep([123, 456])
    assert looptime == 123
    assert unslept is None


async def test_specific_delays_only_are_awaited(looptime):
    unslept = await sleep([123, None])
    assert looptime == 123
    assert unslept is None


@pytest.mark.parametrize('delays', [
    pytest.param([1000, -10], id='mixed-signs'),
    pytest.param([-100, -10], id='all-negative'),
    pytest.param(-10, id='alone'),
])
async def test_negative_delays_skip_sleeping(looptime, delays):
    unslept = await sleep(delays)
    assert looptime == 0
    assert unslept is None


@pytest.mark.parametrize('delays', [
    pytest.param([], id='empty-list'),
    pytest.param([None], id='list-of-none'),
])
async def test_no_delays_skip_sleeping(looptime, delays):
    unslept = await sleep(delays)
    assert looptime == 0
    assert unslept is None


async def test_by_event_set_before_time_comes(looptime):
    event = asyncio.Event()
    asyncio.get_running_loop().call_later(7, event.set)
    unslept = await sleep(10, event)
    assert unslept is not None
    assert unslept == 3
    assert looptime == 7


async def test_with_zero_time_and_event_initially_cleared(looptime):
    event = asyncio.Event()
    event.clear()
    unslept = await sleep(0, event)
    assert looptime == 0
    assert unslept is None


async def test_with_zero_time_and_event_initially_set(looptime):
    event = asyncio.Event()
    event.set()
    unslept = await sleep(0, event)
    assert looptime == 0
    assert not unslept  # 0/None; undefined for such case: both goals reached.



"""
An internal in-memory storage of structured records about resource objects.

Each object gets at most one record in the inventory of an operator's memories.
The information is stored strictly in-memory and is not persistent.
On the operator restart, all the memories are lost.
The information is not exposed to the operator developers (except for memos).

It is used internally to track allocated system resources for each Kubernetes
object, even if that object does not show up in the event streams for long time.

In the future, additional never-ending tasks can be running to maintain
the operator's memories and inventory and garbage-collect all outdated records.

The inventory memories are data structures, but they are a part of the reactor
because they store specialised data structures of specialised reactor modules
(e.g. daemons, admission, indexing, etc). For cohesion, these data structures
must be kept together with their owning modules rather than mirrored in structs.
"""
import copy
import dataclasses
from collections.abc import Iterator, MutableMapping

from kopf._cogs.structs import bodies, ephemera
from kopf._core.actions import throttlers
from kopf._core.engines import admission, daemons, indexing


@dataclasses.dataclass(frozen=False)
class ResourceMemory:
    """
    A system memo about a single resource/object.

    Usually stored in :class:`Memories`.
    """
    memo: ephemera.AnyMemo = dataclasses.field(default_factory=lambda: ephemera.AnyMemo(ephemera.Memo()))
    error_throttler: throttlers.Throttler = dataclasses.field(default_factory=throttlers.Throttler)
    indexing_memory: indexing.IndexingMemory = dataclasses.field(default_factory=indexing.IndexingMemory)
    daemons_memory: daemons.DaemonsMemory = dataclasses.field(default_factory=daemons.DaemonsMemory)

    # For resuming handlers tracking and deciding on should they be called or not.
    noticed_by_listing: bool = False
    fully_handled_once: bool = False


class ResourceMemories(admission.MemoGetter, daemons.DaemonsMemoriesIterator):
    """
    A container of all memos about every existing resource in a single operator.

    Distinct operator tasks have their own memory containers, which
    do not overlap. This solves the problem if storing the per-resource
    entries in the global or context variables.

    The memos can store anything the resource handlers need to persist within
    a single process/operator lifetime, but not persisted on the resource.
    For example, the runtime system resources: flags, threads, tasks, etc.
    Or the scalar values, which have meaning only for this operator process.

    The container is relatively async-safe: one individual resource is always
    handled sequentially, never in parallel with itself (different resources
    are handled in parallel through), so the same key will not be added/deleted
    in the background during the operation, so the locking is not needed.
    """
    _items: MutableMapping[str, ResourceMemory]

    def __init__(self) -> None:
        super().__init__()
        self._items = {}

    def iter_all_memories(self) -> Iterator[ResourceMemory]:
        yield from self._items.values()

    def iter_all_daemon_memories(self) -> Iterator[daemons.DaemonsMemory]:
        for memory in self._items.values():
            yield memory.daemons_memory

    async def recall_memo(
            self,
            raw_body: bodies.RawBody,
            *,
            memobase: ephemera.AnyMemo | None = None,
            ephemeral: bool = False,
    ) -> ephemera.AnyMemo:
        memory = await self.recall(raw_body=raw_body, memobase=memobase, ephemeral=ephemeral)
        return memory.memo

    async def recall(
            self,
            raw_body: bodies.RawBody,
            *,
            memobase: ephemera.AnyMemo | None = None,
            noticed_by_listing: bool = False,
            ephemeral: bool = False,
    ) -> ResourceMemory:
        """
        Either find a resource's memory, or create and remember a new one.

        Keep the last-seen body up to date for all the handlers.

        Ephemeral memos are not remembered now
        (later: will be remembered for short time, and then garbage-collected).
        They are used by admission webhooks before the resource is created --
        to not waste RAM for what might never exist. The persistent memo
        will be created *after* the resource creation really happens.
        """
        key = self._build_key(raw_body)
        if key in self._items:
            memory = self._items[key]
        else:
            if memobase is None:
                memory = ResourceMemory(noticed_by_listing=noticed_by_listing)
            else:
                memo = copy.copy(memobase)
                memory = ResourceMemory(noticed_by_listing=noticed_by_listing, memo=memo)
            if not ephemeral:
                self._items[key] = memory
        return memory

    async def forget(
            self,
            raw_body: bodies.RawBody,
    ) -> None:
        """
        Forget the resource's memory if it exists; or ignore if it does not.
        """
        key = self._build_key(raw_body)
        if key in self._items:
            del self._items[key]

    def _build_key(
            self,
            raw_body: bodies.RawBody,
    ) -> str:
        """
        Construct an immutable persistent key of a resource.

        Generally, a uid is sufficient, as it is unique within the cluster.
        But it can be e.g. plural/namespace/name triplet, or anything else,
        even of different types (as long as it satisfies the type checkers).

        But it must be consistent within a single process lifetime.
        """
        return raw_body.get('metadata', {}).get('uid') or ''



from unittest.mock import Mock

import pytest

import kopf
from kopf._cogs.structs.ids import HandlerId
from kopf._core.engines.admission import serve_admission_request
from kopf._core.intents.causes import WebhookType


async def test_all_handlers_with_no_id_or_reason_requested(
        settings, registry, resource, memories, insights, indices, adm_request):

    mock1 = Mock()
    mock2 = Mock()
    mock3 = Mock()
    mock4 = Mock()

    @kopf.on.validate(*resource)
    def fn1(**kwargs):
        mock1(**kwargs)

    @kopf.on.validate(*resource)
    def fn2(**kwargs):
        mock2(**kwargs)

    @kopf.on.mutate(*resource)
    def fn3(**kwargs):
        mock3(**kwargs)

    @kopf.on.mutate(*resource)
    def fn4(**kwargs):
        mock4(**kwargs)

    response = await serve_admission_request(
        adm_request,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert response['response']['allowed'] is True
    assert mock1.call_count == 1
    assert mock2.call_count == 1
    assert mock3.call_count == 1
    assert mock4.call_count == 1


@pytest.mark.parametrize('reason', set(WebhookType))
async def test_handlers_with_reason_requested(
        settings, registry, resource, memories, insights, indices, adm_request, reason):

    mock1 = Mock()
    mock2 = Mock()
    mock3 = Mock()
    mock4 = Mock()

    @kopf.on.validate(*resource)
    def fn1(**kwargs):
        mock1(**kwargs)

    @kopf.on.validate(*resource)
    def fn2(**kwargs):
        mock2(**kwargs)

    @kopf.on.mutate(*resource)
    def fn3(**kwargs):
        mock3(**kwargs)

    @kopf.on.mutate(*resource)
    def fn4(**kwargs):
        mock4(**kwargs)

    response = await serve_admission_request(
        adm_request, reason=reason,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert response['response']['allowed'] is True
    assert mock1.call_count == (1 if reason == WebhookType.VALIDATING else 0)
    assert mock2.call_count == (1 if reason == WebhookType.VALIDATING else 0)
    assert mock3.call_count == (1 if reason == WebhookType.MUTATING else 0)
    assert mock4.call_count == (1 if reason == WebhookType.MUTATING else 0)


async def test_handlers_with_webhook_requested(
        settings, registry, resource, memories, insights, indices, adm_request):

    mock1 = Mock()
    mock2 = Mock()
    mock3 = Mock()
    mock4 = Mock()

    @kopf.on.validate(*resource, id='fnX')
    def fn1(**kwargs):
        mock1(**kwargs)

    @kopf.on.validate(*resource)
    def fn2(**kwargs):
        mock2(**kwargs)

    @kopf.on.mutate(*resource)
    def fn3(**kwargs):
        mock3(**kwargs)

    @kopf.on.mutate(*resource, id='fnX')
    def fn4(**kwargs):
        mock4(**kwargs)

    response = await serve_admission_request(
        adm_request, webhook=HandlerId('fnX'),
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert response['response']['allowed'] is True
    assert mock1.call_count == 1
    assert mock2.call_count == 0
    assert mock3.call_count == 0
    assert mock4.call_count == 1


@pytest.mark.parametrize('reason', set(WebhookType))
async def test_handlers_with_reason_and_webhook_requested(
        settings, registry, resource, memories, insights, indices, adm_request, reason):

    mock1 = Mock()
    mock2 = Mock()
    mock3 = Mock()
    mock4 = Mock()

    @kopf.on.validate(*resource, id='fnX')
    def fn1(**kwargs):
        mock1(**kwargs)

    @kopf.on.validate(*resource)
    def fn2(**kwargs):
        mock2(**kwargs)

    @kopf.on.mutate(*resource)
    def fn3(**kwargs):
        mock3(**kwargs)

    @kopf.on.mutate(*resource, id='fnX')
    def fn4(**kwargs):
        mock4(**kwargs)

    response = await serve_admission_request(
        adm_request, webhook=HandlerId('fnX'), reason=reason,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert response['response']['allowed'] is True
    assert mock1.call_count == (1 if reason == WebhookType.VALIDATING else 0)
    assert mock2.call_count == 0
    assert mock3.call_count == 0
    assert mock4.call_count == (1 if reason == WebhookType.MUTATING else 0)


@pytest.mark.parametrize('operation', ['CREATE', 'UPDATE', 'CONNECT', '*WHATEVER*'])
async def test_mutating_handlers_are_selected_for_nondeletion(
        settings, registry, resource, memories, insights, indices, adm_request, operation):

    v_mock = Mock()
    m_mock = Mock()

    @kopf.on.validate(*resource)
    def v_fn(**kwargs):
        v_mock(**kwargs)

    @kopf.on.mutate(*resource)
    def m_fn(**kwargs):
        m_mock(**kwargs)

    adm_request['request']['operation'] = operation
    response = await serve_admission_request(
        adm_request,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert response['response']['allowed'] is True
    assert v_mock.call_count == 1
    assert m_mock.call_count == 1


async def test_mutating_handlers_are_not_selected_for_deletion_by_default(
        settings, registry, resource, memories, insights, indices, adm_request):

    v_mock = Mock()
    m_mock = Mock()

    @kopf.on.validate(*resource)
    def v_fn(**kwargs):
        v_mock(**kwargs)

    @kopf.on.mutate(*resource)
    def m_fn(**kwargs):
        m_mock(**kwargs)

    adm_request['request']['operation'] = 'DELETE'
    response = await serve_admission_request(
        adm_request,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert response['response']['allowed'] is True
    assert v_mock.call_count == 1
    assert m_mock.call_count == 0


async def test_mutating_handlers_are_selected_for_deletion_if_explicitly_marked(
        settings, registry, resource, memories, insights, indices, adm_request):

    v_mock = Mock()
    m_mock = Mock()

    @kopf.on.validate(*resource)
    def v_fn(**kwargs):
        v_mock(**kwargs)

    @kopf.on.mutate(*resource, operations=['DELETE'])
    def m_fn(**kwargs):
        m_mock(**kwargs)

    adm_request['request']['operation'] = 'DELETE'
    response = await serve_admission_request(
        adm_request,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert response['response']['allowed'] is True
    assert v_mock.call_count == 1
    assert m_mock.call_count == 1


@pytest.mark.parametrize('handler_sub, request_sub, exp_calls', [
    ('*', 'no', 1),
    ('*', 'sb', 1),
    ('*', None, 1),
    ('sb', 'no', 0),
    ('sb', 'sb', 1),
    ('sb', None, 0),
    (None, 'no', 0),
    (None, 'sb', 0),
    (None, None, 1),
])
async def test_subresources(
        settings, registry, resource, memories, insights, indices, adm_request,
        handler_sub, request_sub, exp_calls):

    v_mock = Mock()
    m_mock = Mock()

    @kopf.on.validate(*resource, subresource=handler_sub)
    def v_fn(**kwargs):
        v_mock(**kwargs)

    @kopf.on.mutate(*resource, subresource=handler_sub)
    def m_fn(**kwargs):
        m_mock(**kwargs)

    adm_request['request']['subResource'] = request_sub
    adm_request['request']['requestSubResource'] = request_sub
    response = await serve_admission_request(
        adm_request,
        settings=settings, registry=registry, insights=insights,
        memories=memories, memobase=object(), indices=indices,
    )
    assert response['response']['allowed'] is True
    assert v_mock.call_count == exp_calls
    assert m_mock.call_count == exp_calls



"""
Just to make sure that asyncio tests are configured properly.
"""
import asyncio

_async_was_executed = False


async def test_async_tests_are_enabled():
    global _async_was_executed
    _async_was_executed = True  # asserted in a sync-test below.


async def test_async_mocks_are_enabled(mocker, looptime):
    p = mocker.patch('asyncio.sleep')
    await asyncio.sleep(1.0)

    assert p.call_count > 0
    assert p.await_count > 0
    assert looptime == 0


def test_async_test_was_executed_and_awaited():
    assert _async_was_executed



import asyncio
import logging

import pytest

from kopf._cogs.aiokits.aiotoggles import ToggleSet
from kopf._cogs.clients.watching import streaming_block


async def test_pausing_is_ignored_if_turned_off(
        resource, namespace, looptime, assert_logs):
    operator_paused = ToggleSet(any)
    await operator_paused.make_toggle(False)

    async with streaming_block(
        resource=resource,
        namespace=namespace,
        operator_paused=operator_paused,
    ):
        pass

    assert looptime == 0
    assert_logs(prohibited=[
        r"Pausing the watch-stream for",
        r"Resuming the watch-stream for",
    ])


async def test_pausing_waits_forever_if_not_resumed(
        resource, namespace, looptime, assert_logs):
    operator_paused = ToggleSet(any)
    await operator_paused.make_toggle(True)

    async def do_it():
        async with streaming_block(
                resource=resource,
                namespace=namespace,
                operator_paused=operator_paused,
        ):
            pass

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(do_it(), timeout=1.23)

    assert looptime == 1.23
    assert_logs([
        r"Pausing the watch-stream for",
    ], prohibited=[
        r"Resuming the watch-stream for",
    ])


async def test_pausing_waits_until_resumed(
        resource, namespace, looptime, assert_logs):
    operator_paused = ToggleSet(any)
    conflicts_found = await operator_paused.make_toggle(True)

    async def delayed_resuming(delay: float):
        await asyncio.sleep(delay)
        await conflicts_found.turn_to(False)

    asyncio.create_task(delayed_resuming(1.23))
    async with streaming_block(
        resource=resource,
        namespace=namespace,
        operator_paused=operator_paused,
    ):
        pass

    assert looptime == 1.23
    assert_logs([
        r"Pausing the watch-stream for",
        r"Resuming the watch-stream for",
    ])



"""
Module- and file-loading to trigger the handlers to be registered.

Since the framework is based on the decorators to register the handlers,
the files/modules with these handlers should be loaded first,
thus executing the decorators.

The files/modules to be loaded are usually specified on the command-line.
Currently, two loading modes are supported, both are equivalent to Python CLI:

* Plain files files (`kopf run file.py`).
* Importable modules (`kopf run -m pkg.mod`).

Multiple files/modules can be specified. They will be loaded in the order.
"""

import importlib
import importlib.abc
import importlib.util
import os.path
import sys
from collections.abc import Iterable
from typing import cast


def preload(
        paths: Iterable[str],
        modules: Iterable[str],
) -> None:
    """
    Ensure the handlers are registered by loading/importing the files/modules.
    """

    for idx, path in enumerate(paths):
        sys.path.insert(0, os.path.abspath(os.path.dirname(path)))
        name = f'__kopf_script_{idx}__{path}'  # same pseudo-name as '__main__'
        spec = importlib.util.spec_from_file_location(name, path)
        module = importlib.util.module_from_spec(spec) if spec is not None else None
        loader = cast(importlib.abc.Loader, spec.loader) if spec is not None else None
        if module is not None and loader is not None:
            sys.modules[name] = module
            loader.exec_module(module)
        else:
            raise ImportError(f"Failed loading {path}: no module or loader.")

    for name in modules:
        importlib.import_module(name)



import pytest

from kopf._cogs.structs.dicts import parse_field


def test_from_none():
    path = parse_field(None)
    assert isinstance(path, tuple)
    assert len(path) == 0


def test_from_string_one_level():
    path = parse_field('field')
    assert isinstance(path, tuple)
    assert path == ('field',)


def test_from_string_two_levels():
    path = parse_field('field.subfield')
    assert isinstance(path, tuple)
    assert path == ('field', 'subfield')


def test_from_list():
    path = parse_field(['field' , 'subfield'])
    assert isinstance(path, tuple)
    assert path == ('field', 'subfield')


def test_from_tuple():
    path = parse_field(('field' , 'subfield'))
    assert isinstance(path, tuple)
    assert path == ('field', 'subfield')


@pytest.mark.parametrize('val', [dict(), set(), frozenset()])
def test_from_others_fails(val):
    with pytest.raises(ValueError):
        parse_field(val)



from typing import cast

from kopf._cogs.clients import api
from kopf._cogs.configs import configuration
from kopf._cogs.helpers import typedefs
from kopf._cogs.structs import bodies, references


async def create_obj(
        *,
        settings: configuration.OperatorSettings,
        resource: references.Resource,
        namespace: references.Namespace = None,
        name: str | None = None,
        body: bodies.RawBody | None = None,
        logger: typedefs.Logger,
) -> bodies.RawBody | None:
    """
    Create a resource.
    """
    body = body if body is not None else {}
    if namespace is not None:
        body.setdefault('metadata', {}).setdefault('namespace', namespace)
    if name is not None:
        body.setdefault('metadata', {}).setdefault('name', name)

    namespace = cast(references.Namespace, body.get('metadata', {}).get('namespace'))
    created_body: bodies.RawBody = await api.post(
        url=resource.get_url(namespace=namespace),
        payload=body,
        logger=logger,
        settings=settings,
    )
    return created_body



"""
Watching and streaming watch-events.

Kubernetes client's watching streams are synchronous. To make them asynchronous,
we put them into a :class:`concurrent.futures.ThreadPoolExecutor`,
and yield from there asynchronously.

However, async/await coroutines misbehave with ``StopIteration`` exceptions
raised by the ``next`` method: see `PEP-479`_.

As a workaround, we replace ``StopIteration`` with our custom ``StopStreaming``
inherited from ``RuntimeError`` (as suggested by `PEP-479`_),
and re-implement the generators to make them async.

All of this is a workaround for the standard Kubernetes client's limitations.
They would not be needed if the client library were natively asynchronous.

.. _PEP-479: https://www.python.org/dev/peps/pep-0479/
"""
import asyncio
import contextlib
import enum
import logging
from collections.abc import AsyncIterator
from typing import cast

import aiohttp

from kopf._cogs.aiokits import aiotasks, aiotoggles
from kopf._cogs.clients import api, errors, fetching
from kopf._cogs.configs import configuration
from kopf._cogs.structs import bodies, references

logger = logging.getLogger(__name__)

HTTP_TOO_MANY_REQUESTS_CODE = 429
DEFAULT_RETRY_DELAY_SECONDS = 1


class WatchingError(Exception):
    """
    Raised when an unexpected error happens in the watch-stream API.
    """


class Bookmark(enum.Enum):
    """ Special marks sent in the stream among raw events. """
    LISTED = enum.auto()  # the listing is over, now streaming.


async def infinite_watch(
        *,
        settings: configuration.OperatorSettings,
        resource: references.Resource,
        namespace: references.Namespace,
        operator_paused: aiotoggles.ToggleSet | None = None,  # None for tests & observation
        _iterations: int | None = None,  # used in tests/mocks/fixtures
) -> AsyncIterator[Bookmark | bodies.RawEvent]:
    """
    Stream the watch-events infinitely.

    This routine is extracted because it is difficult to test infinite loops.
    It is made as simple as possible, and is assumed to work without testing.

    This routine never ends gracefully. If a watcher's stream fails,
    a new one is recreated, and the stream continues.
    It only exits with unrecoverable exceptions.
    """
    how = ' (paused)' if operator_paused is not None and operator_paused.is_on() else ''
    where = f'in {namespace!r}' if namespace is not None else 'cluster-wide'
    logger.debug(f"Starting the watch-stream for {resource} {where}{how}.")
    try:
        while _iterations is None or _iterations > 0:  # equivalent to `while True` in non-test mode
            _iterations = None if _iterations is None else _iterations - 1
            async with streaming_block(
                namespace=namespace,
                resource=resource,
                operator_paused=operator_paused,
            ) as operator_pause_waiter:
                stream = continuous_watch(
                    settings=settings,
                    resource=resource,
                    namespace=namespace,
                    operator_pause_waiter=operator_pause_waiter,
                )
                try:
                    async for raw_event in stream:
                        yield raw_event
                except errors.APIClientError as ex:
                    if ex.code != HTTP_TOO_MANY_REQUESTS_CODE:
                        raise

                    retry_after = ex.details.get("retryAfterSeconds") if ex.details else None
                    retry_wait = retry_after or DEFAULT_RETRY_DELAY_SECONDS
                    logger.warning(
                        f"Receiving `too many requests` error from server, will retry after "
                        f"{retry_wait} seconds. Error details: {ex}"
                    )
                    await asyncio.sleep(retry_wait)
            await asyncio.sleep(settings.watching.reconnect_backoff)
    finally:
        logger.debug(f"Stopping the watch-stream for {resource} {where}.")


@contextlib.asynccontextmanager
async def streaming_block(
        *,
        resource: references.Resource,
        namespace: references.Namespace,
        operator_paused: aiotoggles.ToggleSet | None = None,  # None for tests & observation
) -> AsyncIterator[aiotasks.Future]:
    """
    Block the execution until un-paused; signal when it is active again.

    This prevents both watching and listing while the operator is paused,
    until it is off. Specifically, the watch-stream closes its connection
    once paused, so the while-true & for-event-in-stream cycles exit,
    and the streaming coroutine is started again by ``infinite_stream()``
    (the watcher timeout is swallowed by the pause time).

    Returns a future (or a task) that is set (or finished) when paused again.

    A stop-future is a client-specific way of terminating the streaming HTTPS
    connections when paused again. The low-level streaming API call attaches
    its ``response.close()`` to the future's "done" callback,
    so that the stream is closed once the operator is paused.

    Note: this routine belongs to watching and does not belong to peering.
    The pause can be managed in any other ways: as an imaginary edge case,
    imagine a operator with UI with a "pause" button that pauses the operator.
    """
    where = f'in {namespace!r}' if namespace is not None else 'cluster-wide'

    # Block until unpaused before even starting the API communication.
    if operator_paused is not None and operator_paused.is_on():
        names = {toggle.name for toggle in operator_paused if toggle.is_on() and toggle.name}
        pause_reason = f" (blockers: {', '.join(names)})" if names else ""
        logger.debug(f"Pausing the watch-stream for {resource} {where}{pause_reason}.")

        await operator_paused.wait_for(False)

        names = {toggle.name for toggle in operator_paused if toggle.is_on() and toggle.name}
        resuming_reason = f" (resolved: {', '.join(names)})" if names else ""
        logger.debug(f"Resuming the watch-stream for {resource} {where}{resuming_reason}.")

    # Create the signalling future for when paused again.
    operator_pause_waiter: aiotasks.Future
    if operator_paused is not None:
        operator_pause_waiter = asyncio.create_task(
            operator_paused.wait_for(True),
            name=f"pause-waiter for {resource}")
    else:
        operator_pause_waiter = asyncio.Future()  # a dummy just to have it

    # Go for the streaming with the prepared pauseing/unpausing setup.
    try:
        yield operator_pause_waiter
    finally:
        with contextlib.suppress(asyncio.CancelledError):
            operator_pause_waiter.cancel()
            await operator_pause_waiter


async def continuous_watch(
        *,
        settings: configuration.OperatorSettings,
        resource: references.Resource,
        namespace: references.Namespace,
        operator_pause_waiter: aiotasks.Future,
) -> AsyncIterator[Bookmark | bodies.RawEvent]:

    # First, list the resources regularly, and get the list's resource version.
    # Simulate the events with type "None" event - used in detection of causes.
    try:
        objs, resource_version = await fetching.list_objs(
            logger=logger,
            settings=settings,
            resource=resource,
            namespace=namespace,
        )
        for obj in objs:
            yield {'type': None, 'object': obj}

    except (aiohttp.ClientConnectionError, aiohttp.ClientPayloadError, asyncio.TimeoutError):
        return

    # Notify the watcher that the initial listing is over, even if there was nothing yielded.
    yield Bookmark.LISTED

    # Repeat through disconnects of the watch as long as the resource version is valid (no errors).
    # The individual watching API calls are disconnected by timeout even if the stream is fine.
    while not operator_pause_waiter.done():

        # Then, watch the resources starting from the list's resource version.
        stream = watch_objs(
            settings=settings,
            resource=resource,
            namespace=namespace,
            since=resource_version,
            operator_pause_waiter=operator_pause_waiter,
        )
        async for raw_input in stream:
            raw_type = raw_input['type']
            raw_object = raw_input['object']

            # "410 Gone" is for the "resource version too old" error, we must restart watching.
            # The resource versions are lost by k8s after a few minutes (5 as per the official doc).
            # The error occurs when there is nothing happening for a few minutes. This is normal.
            if raw_type == 'ERROR' and cast(bodies.RawError, raw_object)['code'] == 410:
                where = f'in {namespace!r}' if namespace is not None else 'cluster-wide'
                logger.debug(f"Restarting the watch-stream for {resource} {where}.")
                return  # out of the regular stream, to the infinite stream.

            # Other watch errors should be fatal for the operator.
            if raw_type == 'ERROR':
                raise WatchingError(f"Error in the watch-stream: {raw_object}")

            # Ensure that the event is something we understand and can handle.
            if raw_type not in ['ADDED', 'MODIFIED', 'DELETED']:
                logger.warning(f"Ignoring an unsupported event type: {raw_input!r}")
                continue

            # Keep the latest seen resource version for continuation of the stream on disconnects.
            body = cast(bodies.RawBody, raw_object)
            resource_version = body.get('metadata', {}).get('resourceVersion', resource_version)

            # Yield normal events to the consumer. Errors are already filtered out.
            yield cast(bodies.RawEvent, raw_input)


async def watch_objs(
        *,
        settings: configuration.OperatorSettings,
        resource: references.Resource,
        namespace: references.Namespace,
        since: str | None = None,
        operator_pause_waiter: aiotasks.Future,
) -> AsyncIterator[bodies.RawInput]:
    """
    Watch objects of a specific resource type.

    The cluster-scoped call is used in two cases:

    * The resource itself is cluster-scoped, and namespacing makes not sense.
    * The operator serves all namespaces for the namespaced custom resource.

    Otherwise, the namespace-scoped call is used:

    * The resource is namespace-scoped AND operator is namespaced-restricted.
    """
    params: dict[str, str] = {}
    params['watch'] = 'true'
    if since is not None:
        params['resourceVersion'] = since
    if settings.watching.server_timeout is not None:
        params['timeoutSeconds'] = str(settings.watching.server_timeout)

    connect_timeout = (
        settings.watching.connect_timeout if settings.watching.connect_timeout is not None else
        settings.networking.connect_timeout if settings.networking.connect_timeout is not None else
        settings.networking.request_timeout
    )

    # Stream the parsed events from the response until it is closed server-side,
    # or until it is closed client-side by the pause-waiting future's callbacks.
    try:
        async for raw_input in api.stream(
            url=resource.get_url(namespace=namespace, params=params),
            logger=logger,
            settings=settings,
            stopper=operator_pause_waiter,
            timeout=aiohttp.ClientTimeout(
                total=settings.watching.client_timeout,
                sock_connect=connect_timeout,
            ),
        ):
            yield raw_input

    except (aiohttp.ClientConnectionError, aiohttp.ClientPayloadError, asyncio.TimeoutError):
        pass



import asyncio

from kopf._cogs.aiokits.aiotasks import all_tasks


async def test_alltasks_exclusion():
    flag = asyncio.Event()
    task1 = asyncio.create_task(flag.wait())
    task2 = asyncio.create_task(flag.wait())
    done, pending = await asyncio.wait([task1, task2], timeout=0.01)  # let them start
    assert not done

    tasks = await all_tasks(ignored=[task2])
    assert task1 in tasks
    assert task2 not in tasks
    assert asyncio.current_task() not in tasks

    flag.set()
    await task1
    await task2



import collections.abc

import pytest

from kopf._cogs.structs.diffs import DiffOperation, DiffScope, diff


@pytest.mark.parametrize('scope', list(DiffScope))
def test_none_for_old(scope):
    a = None
    b = object()
    d = diff(a, b, scope=scope)
    assert d == (('add', (), None, b),)


@pytest.mark.parametrize('scope', list(DiffScope))
def test_none_for_new(scope):
    a = object()
    b = None
    d = diff(a, b, scope=scope)
    assert d == (('remove', (), a, None),)


@pytest.mark.parametrize('scope', list(DiffScope))
def test_nones_for_both(scope):
    a = None
    b = None
    d = diff(a, b, scope=scope)
    assert d == ()


@pytest.mark.parametrize('scope', list(DiffScope))
def test_scalars_equal(scope):
    a = 100
    b = 100
    d = diff(a, b, scope=scope)
    assert d == ()


@pytest.mark.parametrize('scope', list(DiffScope))
def test_scalars_unequal(scope):
    a = 100
    b = 200
    d = diff(a, b, scope=scope)
    assert d == (('change', (), 100, 200),)


@pytest.mark.parametrize('scope', list(DiffScope))
def test_strings_equal(scope):
    a = 'hello'
    b = 'hello'
    d = diff(a, b, scope=scope)
    assert d == ()


@pytest.mark.parametrize('scope', list(DiffScope))
def test_strings_unequal(scope):
    a = 'hello'
    b = 'world'
    d = diff(a, b, scope=scope)
    assert d == (('change', (), 'hello', 'world'),)


@pytest.mark.parametrize('scope', list(DiffScope))
def test_lists_equal(scope):
    a = [100, 200, 300]
    b = [100, 200, 300]
    d = diff(a, b, scope=scope)
    assert d == ()


@pytest.mark.parametrize('scope', list(DiffScope))
def test_lists_unequal(scope):
    a = [100, 200, 300]
    b = [100, 666, 300]
    d = diff(a, b, scope=scope)
    assert d == (('change', (), [100, 200, 300], [100, 666, 300]),)


@pytest.mark.parametrize('scope', list(DiffScope))
def test_dicts_equal(scope):
    a = {'hello': 'world', 'key': 'val'}
    b = {'key': 'val', 'hello': 'world'}
    d = diff(a, b, scope=scope)
    assert d == ()


@pytest.mark.parametrize('scope', [DiffScope.FULL, DiffScope.RIGHT])
def test_dicts_with_keys_added_and_noticed(scope):
    a = {'hello': 'world'}
    b = {'hello': 'world', 'key': 'val'}
    d = diff(a, b, scope=scope)
    assert d == (('add', ('key',), None, 'val'),)


@pytest.mark.parametrize('scope', [DiffScope.LEFT])
def test_dicts_with_keys_added_but_ignored(scope):
    a = {'hello': 'world'}
    b = {'hello': 'world', 'key': 'val'}
    d = diff(a, b, scope=scope)
    assert d == ()


@pytest.mark.parametrize('scope', [DiffScope.FULL, DiffScope.LEFT])
def test_dicts_with_keys_removed_and_noticed(scope):
    a = {'hello': 'world', 'key': 'val'}
    b = {'hello': 'world'}
    d = diff(a, b, scope=scope)
    assert d == (('remove', ('key',), 'val', None),)


@pytest.mark.parametrize('scope', [DiffScope.RIGHT])
def test_dicts_with_keys_removed_but_ignored(scope):
    a = {'hello': 'world', 'key': 'val'}
    b = {'hello': 'world'}
    d = diff(a, b, scope=scope)
    assert d == ()


@pytest.mark.parametrize('scope', list(DiffScope))
def test_dicts_with_keys_changed(scope):
    a = {'hello': 'world', 'key': 'old'}
    b = {'hello': 'world', 'key': 'new'}
    d = diff(a, b, scope=scope)
    assert d == (('change', ('key',), 'old', 'new'),)


@pytest.mark.parametrize('scope', list(DiffScope))
def test_dicts_with_subkeys_changed(scope):
    a = {'main': {'hello': 'world', 'key': 'old'}}
    b = {'main': {'hello': 'world', 'key': 'new'}}
    d = diff(a, b, scope=scope)
    assert d == (('change', ('main', 'key'), 'old', 'new'),)


def test_custom_mappings_are_recursed():

    class SampleMapping(collections.abc.Mapping):
        def __init__(self, data=(), **kwargs) -> None:
            super().__init__()
            self._items = dict(data) | kwargs
        def __len__(self) -> int: return len(self._items)
        def __iter__(self): return iter(self._items)
        def __getitem__(self, item: str) -> str: return self._items[item]

    class MappingA(SampleMapping): pass
    class MappingB(SampleMapping): pass

    a = MappingA(a=100, b=200)
    b = MappingB(b=300, c=400)
    d = diff(a, b)
    assert (DiffOperation.REMOVE, ('a',), 100, None) in d
    assert (DiffOperation.CHANGE, ('b',), 200, 300) in d
    assert (DiffOperation.ADD, ('c',), None, 400) in d
    assert (DiffOperation.CHANGE, (), a, b) not in d


# A few examples of slightly more realistic non-abstracted use-cases below:
def test_dicts_adding_label():
    body_before_labelling = {'metadata': {}}
    body_after_labelling  = {'metadata': {'labels': 'LABEL'}}

    d = diff(body_before_labelling, body_after_labelling)
    assert d == (('add', ('metadata', 'labels'), None, 'LABEL'),)


def test_dicts_updating_storage_size():
    body_before_storage_size_update = {'spec': {'size': '42G'}}
    body_after_storage_size_update  = {'spec': {'size': '76G'}}

    d = diff(body_before_storage_size_update, body_after_storage_size_update)
    assert d == (('change', ('spec', 'size'), '42G', '76G'),)


def test_dicts_different_items_handled():
    body_before_storage_size_update = {'spec': {'items': ['task1', 'task2']}}
    body_after_storage_size_update  = {'spec': {'items': ['task3', 'task4']}}

    d = diff(body_before_storage_size_update, body_after_storage_size_update)
    assert d == (('change', ('spec', 'items'), ['task1', 'task2'], ['task3', 'task4']),)



"""
All the functions to write the Kubernetes events for the Kubernetes objects.

They are used internally in the handling routines to show the progress,
and can be used directly from the handlers to add arbitrary custom events.

The actual k8s-event posting runs in the background,
and posts the k8s-events as soon as they are queued.

The k8s-events are queued in two ways:

* Explicit calls to event-posting functions :func:`kopf.event`,
  :func:`kopf.info`, :func:`kopf.warn`, :func:`kopf.exception`.
* Logging messages made on the object logger (above the INFO level by default).

This also includes all logging messages posted by the framework itself.
"""
import asyncio
import logging
import sys
from collections.abc import Iterable, Iterator
from contextvars import ContextVar
from typing import TYPE_CHECKING, NamedTuple, NoReturn, cast

from kopf._cogs.clients import events
from kopf._cogs.configs import configuration
from kopf._cogs.structs import bodies, dicts, references
from kopf._core.actions import loggers

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    K8sEventQueue = asyncio.Queue["K8sEvent"]
else:
    K8sEventQueue = asyncio.Queue

# Logging and event-posting can happen cross-thread: e.g. in sync-executors.
# We have to remember our main event-loop with the queue consumer, to make
# thread-safe coro calls both from inside that event-loop and from outside.
event_queue_loop_var: ContextVar[asyncio.AbstractEventLoop] = ContextVar('event_queue_loop_var')
event_queue_var: ContextVar[K8sEventQueue] = ContextVar('event_queue_var')

# Per-operator container for settings. We only need a log level from there.
# This variable is dedicated to a posting engine, as the call chain is interrupted
# by user-side handlers (no pass-through `settings` arg).
settings_var: ContextVar[configuration.OperatorSettings] = ContextVar('settings_var')


class K8sEvent(NamedTuple):
    """
    A single k8s-event to be posted, with all ref-information preserved.
    It can exist and be posted even after the object is garbage-collected.
    """
    ref: bodies.ObjectReference
    type: str
    reason: str
    message: str


def enqueue(
        ref: bodies.ObjectReference,
        type: str,
        reason: str,
        message: str,
) -> None:
    loop = event_queue_loop_var.get()
    queue = event_queue_var.get()
    event = K8sEvent(ref=ref, type=type, reason=reason, message=message)

    # Events can be posted from another thread than the event-loop's thread
    # (e.g. from sync-handlers, or from explicitly started per-object threads),
    # or from the same thread (async-handlers and the framework itself).
    running_loop: asyncio.AbstractEventLoop | None
    try:
        running_loop = asyncio.get_running_loop()
    except RuntimeError:
        running_loop = None

    if running_loop is loop:
        # Posting from the same event-loop as the poster task and queue are in.
        # Therefore, it is the same thread, and all calls here are thread-safe.
        # Special thread-safe cross-event-loop methods make no effect here.
        queue.put_nowait(event)
    else:
        # No event-loop or another event-loop - assume another thread.
        # Use the cross-thread thread-safe methods. Do not block or wait.
        # Beware of #1212: `run_coroutine_threadsafe(queue.put(…), loop=loop)` is flawed.
        queue.put_nowait(event)
        loop.call_soon_threadsafe(_no_op_event_loop_awakener)


# The same as `lambda: None`, but with no closure data attached & better for JIT in PyPy.
def _no_op_event_loop_awakener() -> None:
    pass


def event(
        objs: bodies.Body | Iterable[bodies.Body],
        *,
        type: str,
        reason: str,
        message: str = '',
) -> None:
    settings: configuration.OperatorSettings = settings_var.get()
    if settings.posting.enabled:
        for obj in cast(Iterator[bodies.Body], dicts.walk(objs)):
            ref = bodies.build_object_reference(obj)
            enqueue(ref=ref, type=type, reason=reason, message=message)


def info(
        objs: bodies.Body | Iterable[bodies.Body],
        *,
        reason: str,
        message: str = '',
) -> None:
    settings: configuration.OperatorSettings = settings_var.get()
    if settings.posting.enabled and settings.posting.level <= logging.INFO:
        for obj in cast(Iterator[bodies.Body], dicts.walk(objs)):
            ref = bodies.build_object_reference(obj)
            enqueue(ref=ref, type='Normal', reason=reason, message=message)


def warn(
        objs: bodies.Body | Iterable[bodies.Body],
        *,
        reason: str,
        message: str = '',
) -> None:
    settings: configuration.OperatorSettings = settings_var.get()
    if settings.posting.level <= logging.WARNING:
        for obj in cast(Iterator[bodies.Body], dicts.walk(objs)):
            ref = bodies.build_object_reference(obj)
            enqueue(ref=ref, type='Warning', reason=reason, message=message)


def exception(
        objs: bodies.Body | Iterable[bodies.Body],
        *,
        reason: str = '',
        message: str = '',
        exc: BaseException | None = None,
) -> None:
    if exc is None:
        _, exc, _ = sys.exc_info()
    reason = reason if reason else type(exc).__name__
    message = f'{message} {exc}' if message and exc else f'{exc}' if exc else f'{message}'
    settings: configuration.OperatorSettings = settings_var.get()
    if settings.posting.enabled and settings.posting.level <= logging.ERROR:
        for obj in cast(Iterator[bodies.Body], dicts.walk(objs)):
            ref = bodies.build_object_reference(obj)
            enqueue(ref=ref, type='Error', reason=reason, message=message)


async def poster(
        *,
        event_queue: K8sEventQueue,
        backbone: references.Backbone,
        settings: configuration.OperatorSettings,
) -> NoReturn:
    """
    Post events in the background as they are queued.

    When the events come from the logging system, they have
    their reason, type, and other fields adjusted to meet Kubernetes's concepts.

    When the events are explicitly defined via :func:`kopf.event` and similar
    calls, they have these special fields defined already.

    In either case, we pass the queued events directly to the K8s client
    (or a client wrapper/adapter), with no extra processing.

    This task is defined in this module only because all other tasks are here,
    so we keep all forever-running tasks together.
    """
    resource = await backbone.wait_for(references.EVENTS)
    while True:
        posted_event = await event_queue.get()
        await events.post_event(
            ref=posted_event.ref,
            type=posted_event.type,
            reason=posted_event.reason,
            message=posted_event.message,
            resource=resource,
            settings=settings,
            logger=logger,
        )


class K8sPoster(logging.Handler):
    """
    A handler to post all log messages as K8s events.
    """
    if sys.version_info[:2] < (3, 13):
        # Disable this optimisation for Python >= 3.13.
        # The `handle` no longer supports having `None` as lock.
        def createLock(self) -> None:
            # Save some time on unneeded locks. Events are posted in the background.
            # We only put events to the queue, which is already lock-protected.
            self.lock = None

    def filter(self, record: logging.LogRecord) -> bool:
        # Only those which have a k8s object referred (see: `ObjectLogger`).
        # Otherwise, we have nothing to post, and nothing to do.
        # TODO: remove all bool() -- they were needed for Python 3.12 & MyPy 1.8.0 wrong inference.
        settings: configuration.OperatorSettings | None
        settings = getattr(record, 'settings', None)
        level_ok = settings is not None and bool(record.levelno >= settings.posting.level)
        enabled = settings is not None and bool(settings.posting.enabled)
        loggers = settings is not None and bool(settings.posting.loggers)
        has_ref = hasattr(record, 'k8s_ref')
        skipped = hasattr(record, 'k8s_skip') and bool(getattr(record, 'k8s_skip'))
        return enabled and level_ok and loggers and has_ref and not skipped and bool(super().filter(record))

    def emit(self, record: logging.LogRecord) -> None:
        # Same try-except as in e.g. `logging.StreamHandler`.
        try:
            ref = getattr(record, 'k8s_ref')
            type = (
                "Debug" if record.levelno <= logging.DEBUG else
                "Normal" if record.levelno <= logging.INFO else
                "Warning" if record.levelno <= logging.WARNING else
                "Error" if record.levelno <= logging.ERROR else
                "Fatal" if record.levelno <= logging.FATAL else
                logging.getLevelName(record.levelno).capitalize())
            reason = 'Logging'
            message = self.format(record)
            enqueue(
                ref=ref,
                type=type,
                reason=reason,
                message=message)
        except Exception:
            self.handleError(record)


loggers.logger.addHandler(K8sPoster())
